@@ -30,7 +30,7 @@ logs = print
 
 
 def generate_classification_from_mosaics(file_10m: Path, file_20m: Path, workspace: Path, training: Path,
-                                         classes: List[int]) -> Iterable[Path]:
+                                         classes: List[int], minimal_support=100) -> Iterable[Path]:
     """
     Generate classification results at 10 and 20 meters for both domains A and B from S2 mosaics (GeoTIFF format)
 
@@ -44,28 +44,35 @@ def generate_classification_from_mosaics(file_10m: Path, file_20m: Path, workspa
         the absolute path to the training_config configuration file
     :param classes: List[int]
         the list of classes to extract from data
+    :param minimal_support: int
+        the minimal support value can be any integer (default is 100)
 
     :return: Iterable[Path]
         the complete path to all classified results saved on disk
     """
 
     cl_a_10m_file, phi_a_10m_file = generate_class(file_10m=file_10m, file_20m=file_20m, output=workspace,
-                                                   training=training, classes=classes, suffix='A', pixres=10)
+                                                   training=training, classes=classes, minimal_support=minimal_support,
+                                                   suffix='A', pixres=10)
 
     cl_b_10m_file, phi_b_10m_file = generate_class(file_10m=file_10m, file_20m=file_20m, output=workspace,
-                                                   training=training, classes=classes, suffix='B', pixres=10)
+                                                   training=training, classes=classes, minimal_support=minimal_support,
+                                                   suffix='B', pixres=10)
 
     cl_a_20m_file, phi_a_20m_file = generate_class(file_10m=file_10m, file_20m=file_20m, output=workspace,
-                                                   training=training, classes=classes, suffix='A', pixres=20)
+                                                   training=training, classes=classes, minimal_support=minimal_support,
+                                                   suffix='A', pixres=20)
 
     cl_b_20m_file, phi_b_20m_file = generate_class(file_10m=file_10m, file_20m=file_20m, output=workspace,
-                                                   training=training, classes=classes, suffix='B', pixres=20)
+                                                   training=training, classes=classes, minimal_support=minimal_support,
+                                                   suffix='B', pixres=20)
 
     return (cl_a_10m_file, phi_a_10m_file, cl_b_10m_file, phi_b_10m_file,
             cl_a_20m_file, phi_a_20m_file, cl_b_20m_file, phi_b_20m_file)
 
 
-def generate_classification_from_safe(filesafe: Path, workspace: Path, training: Path, classes: List[int]) -> Iterable[Path]:
+def generate_classification_from_safe(filesafe: Path, workspace: Path, training: Path, classes: List[int],
+                                      minimal_support=100) -> Iterable[Path]:
     """
     Generate classification results at 10 and 20 meters for both domains A and B from a S2 input (.SAFE or .zip)
 
@@ -77,29 +84,35 @@ def generate_classification_from_safe(filesafe: Path, workspace: Path, training:
         the absolute path to the training_config configuration file
     :param classes: List[int]
         the list of classes to extract from data
+    :param minimal_support: int
+        the minimal support value can be any integer (default is 100)
 
     :return: Iterable[Path]
         the complete path to all classified results saved on disk
     """
 
     cl_a_10m_file, phi_a_10m_file = generate_class_from_safe(filesafe=filesafe, output=workspace, training=training,
-                                                             classes=classes, suffix='A', pixres=10)
+                                                             classes=classes, minimal_support=minimal_support,
+                                                             suffix='A', pixres=10)
 
     cl_b_10m_file, phi_b_10m_file = generate_class_from_safe(filesafe=filesafe, output=workspace, training=training,
-                                                             classes=classes, suffix='B', pixres=10)
+                                                             classes=classes, minimal_support=minimal_support,
+                                                             suffix='B', pixres=10)
 
     cl_a_20m_file, phi_a_20m_file = generate_class_from_safe(filesafe=filesafe, output=workspace, training=training,
-                                                             classes=classes, suffix='A', pixres=20)
+                                                             classes=classes, minimal_support=minimal_support,
+                                                             suffix='A', pixres=20)
 
     cl_b_20m_file, phi_b_20m_file = generate_class_from_safe(filesafe=filesafe, output=workspace, training=training,
-                                                             classes=classes, suffix='B', pixres=20)
+                                                             classes=classes, minimal_support=minimal_support,
+                                                             suffix='B', pixres=20)
 
     return (cl_a_10m_file, phi_a_10m_file, cl_b_10m_file, phi_b_10m_file,
             cl_a_20m_file, phi_a_20m_file, cl_b_20m_file, phi_b_20m_file)
 
 
-def generate_class_from_safe(filesafe: Path, output: Path, training: Path, classes: List[int], suffix: str, pixres: int,
-                             ) -> (Path, Path):
+def generate_class_from_safe(filesafe: Path, output: Path, training: Path, classes: List[int],
+                             minimal_support: int, suffix: str, pixres: int) -> (Path, Path):
     """
     Generate classification results at a given pixel resolution and for a given domain from a S2 input (.SAFE or .zip)
 
@@ -111,6 +124,8 @@ def generate_class_from_safe(filesafe: Path, output: Path, training: Path, class
         the absolute path to the training_config configuration file
     :param classes: List[int]
         the list of classes to extract from data
+    :param minimal_support: int
+        the minimal support value
     :param suffix: str
         the letter of the processing domain: A or B
     :param pixres: int
@@ -127,7 +142,8 @@ def generate_class_from_safe(filesafe: Path, output: Path, training: Path, class
         vrt_20m_file = read_s2_bands_as_vrt(safe_file=filesafe, pixres=20, output=tmp)
 
         class_file, class_phi_file = generate_class(file_10m=vrt_10m_file, file_20m=vrt_20m_file, output=tmp,
-                                                    training=training, classes=classes, suffix=suffix, pixres=pixres)
+                                                    training=training, classes=classes, minimal_support=minimal_support,
+                                                    suffix=suffix, pixres=pixres)
 
         # Move results to workspace
         class_file_dst = output / class_file.name
@@ -185,7 +201,7 @@ def read_s2_bands_as_vrt(safe_file: Path, pixres: int, output: Path) -> Path:
 
 
 def generate_class(file_10m: Path, file_20m: Path, output: Path, training: Path, classes: List[int],
-                   suffix: str, pixres: int) -> (Path, Path):
+                   suffix: str, pixres: int, minimal_support=100) -> (Path, Path):
     """
     Generate classification results at a given pixel resolution and at given domain from S2 mosaics (GeoTIFF format)
 
@@ -199,6 +215,8 @@ def generate_class(file_10m: Path, file_20m: Path, output: Path, training: Path,
         the absolute path to the training_config configuration file
     :param classes: List[int]
         the list of classes to extract from data
+    :param minimal_support: int
+        the minimal support value can be any integer (default is 100)
     :param suffix: str
         the letter of the processing domain: A or B
     :param pixres: int
@@ -222,7 +240,8 @@ def generate_class(file_10m: Path, file_20m: Path, output: Path, training: Path,
         domain = split_domain(file_10m=file_10m, file_20m=file_20m, suffix=suffix, pixres=pixres)
 
         class_file, class_phi_file = process_domain(datafile=main_vrt, suffix=suffix, domain=domain,
-                                                    training=training, classes=classes, output=tmp)
+                                                    training=training, classes=classes, minimal_support=minimal_support,
+                                                    output=tmp)
 
         # Move results to output folder
         class_file_dst = output / class_file.name
@@ -318,7 +337,7 @@ def threshold_otsu(image: np.ndarray) -> int:
 
 
 def process_domain(datafile: Path, suffix: str, domain: np.ndarray, training: Path, classes: List[int],
-                   output: Path) -> (Path, Path):
+                   minimal_support: int, output: Path) -> (Path, Path):
     """
     Produce classification results for a given domain
 
@@ -332,6 +351,8 @@ def process_domain(datafile: Path, suffix: str, domain: np.ndarray, training: Pa
         the absolute path to the training_config configuration file
     :param classes: List[int]
         the list of classes to extract from data
+    :param minimal_support: int
+        the minimal support value
 
     :param output: Path
         the complete path where to write results
@@ -350,7 +371,7 @@ def process_domain(datafile: Path, suffix: str, domain: np.ndarray, training: Pa
     domain_minsupp, datastack_mulquan = sml_minimal_support_multiple_quantization(
         datafile=dataquant_file,
         levels=256,
-        minimal_support=100,
+        minimal_support=minimal_support,
         multiple_quantization=quantizations,
     )
 
@@ -361,7 +382,7 @@ def process_domain(datafile: Path, suffix: str, domain: np.ndarray, training: Pa
 
     logs('Reconciling to a discrete CLASS')
     out_class_file, out_class_phi_file = search_maxima(
-        filename=datastack_class_file, domain_valid=domain, levels=256, output=output,
+        filename=datastack_class_file, domain_valid=domain, levels=200, output=output,
     )
 
     return out_class_file, out_class_phi_file
@@ -651,9 +672,6 @@ def s2_multiple_classification(datafile: Path, suffix: str, domain_valid: np.nda
         height=height,
     )
 
-    sigma = 0.5
-    trunc = np.ceil(2 * sigma) / sigma
-
     for cl, layer in zip(classes, layers):
         logs(f'produce inference for abstraction class with codes: {cl}')
         train_mask = np.isin(train_cgls, cl)
@@ -668,11 +686,9 @@ def s2_multiple_classification(datafile: Path, suffix: str, domain_valid: np.nda
         else:
             phi = np.full(shape=domain_valid.shape, fill_value=np.nan)
 
-        sc_phi = gaussian_filter(phi, sigma=sigma, truncate=trunc)
-
         logs(f'save to file: {layer}')
         with rasterio.open(layer, 'w', **profile) as band:
-            band.write(sc_phi, 1)
+            band.write(phi, 1)
 
     datastack_class_file = output / f'{datafile.stem}_dom{suffix}_sml.vrt'
     gdal.BuildVRT(
@@ -808,10 +824,15 @@ def search_maxima(filename: Path, domain_valid: np.ndarray, levels: int, output:
             count=1,
         )
 
+        profile_phi = profile.copy()
+        profile_phi.update(
+            nodata=255,
+        )
+
         out_class = output / f'{filename.stem}_LC.tif'
         out_class_phi = output / f'{filename.stem}_LC_phi.tif'
         with rasterio.open(out_class, 'w', **profile) as dst_c:
-            with rasterio.open(out_class_phi, 'w', **profile) as dst_phi:
+            with rasterio.open(out_class_phi, 'w', **profile_phi) as dst_phi:
                 for ji, block in src.block_windows(1):
                     # get data and nodata mask by window block
                     data = src.read(window=block)
@@ -822,10 +843,10 @@ def search_maxima(filename: Path, domain_valid: np.ndarray, levels: int, output:
                     warnings.filterwarnings(action='ignore', message='All-NaN slice encountered')
                     phi = np.nanmax(data, axis=0)
                     warnings.resetwarnings()
-                    # rescale in uint8 while keeping 0 for nodata
-                    phi = np.round(imrscl(data=phi, min_value=-1, max_value=1) * (levels - 2)) + 1
-                    phi[np.isnan(phi)] = 0
-                    phi[block_mask] = 0
+                    # rescale in uint8 while keeping 255 for nodata
+                    phi = np.round(imrscl(data=phi, min_value=-1, max_value=1) * levels)
+                    phi[np.isnan(phi)] = 255
+                    phi[block_mask] = 255
 
                     dst_phi.write(phi.astype(np.uint8), 1, window=block)
 
@@ -842,7 +863,7 @@ def search_maxima(filename: Path, domain_valid: np.ndarray, levels: int, output:
     return out_class, out_class_phi
 
 
-def generate_composites(files_10m: List[Path], files_20m: List[Path], output: Path) -> Iterable[Path]:
+def generate_composites(files_10m: List[Path], files_20m: List[Path], output: Path, threshold_phi=0) -> Iterable[Path]:
     """
     Combine all classification results into several composites
 
@@ -853,6 +874,9 @@ def generate_composites(files_10m: List[Path], files_20m: List[Path], output: Pa
 
     :param output: Path
         the complete path where to write results
+    :param threshold_phi: float
+        the minimum phi value to consider the classification valid.
+        It can be any float value between -1 and 1 (default is 0)
 
     :return: Iterable[Path]
         the complete path to all composite results saved on disk
@@ -867,6 +891,7 @@ def generate_composites(files_10m: List[Path], files_20m: List[Path], output: Pa
         pixres=20,
         bounds=bounds,
         output=output,
+        threshold_phi=threshold_phi,
     )
 
     logs('Upsample data 20m to 10m')
@@ -878,7 +903,7 @@ def generate_composites(files_10m: List[Path], files_20m: List[Path], output: Pa
     logs('Upsample phi 20m to 10m')
     composite_20m_to_10m_phi = upsampling_20m_to_10m(
         filename=composite_20m_phi,
-        resampling='nearest',
+        resampling='bilinear',
     )
 
     # get bounds of data resampled from 20m to 10m as they might be slightly larger due to resampling
@@ -891,6 +916,7 @@ def generate_composites(files_10m: List[Path], files_20m: List[Path], output: Pa
         pixres=10,
         bounds=bounds_20m,
         output=output,
+        threshold_phi=threshold_phi,
     )
 
     logs('Merge in composite ALL')
@@ -899,6 +925,7 @@ def generate_composites(files_10m: List[Path], files_20m: List[Path], output: Pa
         comp10m_phi=composite_10m_phi,
         comp20m_data=composite_20m_to_10m,
         comp20m_phi=composite_20m_to_10m_phi,
+        threshold_phi=threshold_phi,
     )
 
     return (composite_20m, composite_20m_phi, composite_20m_count,
@@ -944,7 +971,8 @@ def common_extent_mollweide(filenames: List[Path]) -> BoundingBox:
     return envelope
 
 
-def generate_composite(tiffiles: List[Path], pixres: int, bounds: BoundingBox, output: Path) -> (Path, Path):
+def generate_composite(tiffiles: List[Path], pixres: int, bounds: BoundingBox, output: Path, threshold_phi: float
+                       ) -> (Path, Path):
     """
 
     :param tiffiles: List[Path]
@@ -953,6 +981,8 @@ def generate_composite(tiffiles: List[Path], pixres: int, bounds: BoundingBox, o
         the pixel resolution to filter Sentinel 2 bands, it can be 10 or 20 (meters)
     :param bounds: BoundingBox
         the target spatial extent used to clip
+    :param threshold_phi: float
+        the minimum phi value to consider the classification valid. It can be any float value between -1 and 1
 
     :param output: Path
         the complete path where to write results
@@ -976,22 +1006,43 @@ def generate_composite(tiffiles: List[Path], pixres: int, bounds: BoundingBox, o
         'width': width,
     }
 
-    # init data
-    data = np.zeros(shape=(height, width), dtype=np.uint8)
-    data_phi = np.zeros(shape=(height, width), dtype=np.uint8)
-    data_count = np.zeros(shape=(height, width), dtype=np.uint16)
-
     # filter class and phi files
     files_phi = [f for f in tiffiles if '_phi' in str(f)]
     files_class = [f.parent / f.name.replace('_phi', '') for f in files_phi]
 
+    sigma = 0.5
+    trunc = np.ceil(2 * sigma) / sigma
+
+    # init data
+    with rasterio.open(files_phi[0]) as src_phi:
+        with WarpedVRT(src_phi, **vrt_options) as vrt_phi:
+            nodata = vrt_phi.nodata
+            data_phi = vrt_phi.read(1)
+            nodata_mask = data_phi == nodata
+            # smooth the phi values with gaussian filter
+            data_phi[nodata_mask] = 0
+            data_phi = gaussian_filter(data_phi, sigma=sigma, truncate=trunc)
+            # set back nodata
+            data_phi[nodata_mask] = nodata
+
+    with rasterio.open(files_class[0]) as src:
+        with WarpedVRT(src, **vrt_options) as vrt:
+            data = vrt.read(1)
+
+    data_count = np.zeros(shape=(height, width), dtype=np.uint16)
+    data_count += data_phi != nodata
+
     with rasterio.Env(GDAL_CACHEMAX=512):
-        for fileclass, filephi in zip(files_class, files_phi):
+        for fileclass, filephi in zip(files_class[1:], files_phi[1:]):
 
             # find maximum values and indexes for phi values
             with rasterio.open(filephi) as src_phi:
                 with WarpedVRT(src_phi, **vrt_options) as vrt_phi:
                     next_phi = vrt_phi.read(1)
+                    nodata_mask = next_phi == nodata
+                    # smooth the phi values with gaussian filter
+                    next_phi[nodata_mask] = 0
+                    next_phi = gaussian_filter(next_phi, sigma=sigma, truncate=trunc)
                     better_phi_domain = next_phi > data_phi
 
                     data_phi[better_phi_domain] = next_phi[better_phi_domain]
@@ -1006,6 +1057,14 @@ def generate_composite(tiffiles: List[Path], pixres: int, bounds: BoundingBox, o
             # count amount of data used for each pixel
             data_count += better_phi_domain
 
+        # apply minimum threshold (converted to uint8 value range)
+        threshold_phi_uint8 = np.round(np.interp(threshold_phi, (-1, 1), (0, 255)))
+        # data below the threshold is set to 0 (the "don't know" value)
+        nodata_index = data_phi < threshold_phi_uint8
+        data_phi[nodata_index] = 0
+        data[nodata_index] = 0
+        data_count[nodata_index] = 0
+
         # Write output file
         profile = DefaultGTiffProfile(
             count=1,
@@ -1013,23 +1072,28 @@ def generate_composite(tiffiles: List[Path], pixres: int, bounds: BoundingBox, o
             height=height,
             crs='ESRI:54009',
             transform=transform,
+            nodata=None,
         )
 
+        # write data count
+        composite_count = output / f'composite_S2_CLASS_{pixres}m_stack_count.tif'
+        with rasterio.open(composite_count, 'w', **profile) as dst:
+            dst.write(data_count, 1)
+
         # write data
         composite_data = output / f'composite_S2_CLASS_{pixres}m.tif'
         with rasterio.open(composite_data, 'w', **profile) as dst:
             dst.write(data, 1)
 
         # write phi
+        profile_phi = profile.copy()
+        profile_phi.update(
+            nodata=255,
+        )
         composite_phi = output / f'composite_S2_CLASS_{pixres}m_phi.tif'
-        with rasterio.open(composite_phi, 'w', **profile) as dst:
+        with rasterio.open(composite_phi, 'w', **profile_phi) as dst:
             dst.write(data_phi, 1)
 
-        # write data count
-        composite_count = output / f'composite_S2_CLASS_{pixres}m_stack_count.tif'
-        with rasterio.open(composite_count, 'w', **profile) as dst:
-            dst.write(data_count, 1)
-
     return composite_data, composite_phi, composite_count
 
 
@@ -1058,7 +1122,9 @@ def upsampling_20m_to_10m(filename: Path, resampling: str) -> Path:
     return outfile
 
 
-def generate_composite_all(comp10m_data: Path, comp10m_phi: Path, comp20m_data: Path, comp20m_phi: Path) -> (Path, Path):
+def generate_composite_all(comp10m_data: Path, comp10m_phi: Path,
+                           comp20m_data: Path, comp20m_phi: Path,
+                           threshold_phi: float) -> (Path, Path):
     """
     Blend 10 and 20 meters composite to generate a new one
 
@@ -1070,6 +1136,8 @@ def generate_composite_all(comp10m_data: Path, comp10m_phi: Path, comp20m_data:
         the complete path to the composite of the classification at 20 meter pixel resolution
     :param comp20m_phi: Path
         the complete path to the composite of the phi values at 20 meter pixel resolution
+    :param threshold_phi: float
+        the minimum phi value to consider the classification valid. It can be any float value between -1 and 1
 
     :return: (Path, Path)
          the complete path to the composites of the classification file and the phi value file
@@ -1077,12 +1145,18 @@ def generate_composite_all(comp10m_data: Path, comp10m_phi: Path, comp20m_data:
 
     # read and compare phi
     with rasterio.open(comp10m_phi) as c10m_phi:
-        profile = c10m_phi.profile
+        profile = c10m_phi.profile.copy()
         x_phi = c10m_phi.read()
 
     with rasterio.open(comp20m_phi) as c20m_phi:
         next_phi = c20m_phi.read()
 
+    # apply minimum threshold (converted to uint8 value range)
+    threshold_phi_uint8 = np.round(np.interp(threshold_phi, (-1, 1), (0, 255)))
+    # data below the threshold is set to 0 (the "don't know" value)
+    nodata_index = next_phi < threshold_phi_uint8
+    next_phi[nodata_index] = 0
+
     better_phi_domain = next_phi > x_phi
     x_phi[better_phi_domain] = next_phi[better_phi_domain]
 
@@ -1096,6 +1170,10 @@ def generate_composite_all(comp10m_data: Path, comp10m_phi: Path, comp20m_data:
     x_data[better_phi_domain] = next_data[better_phi_domain]
 
     # Write output file
+    profile.update(
+        nodata=None,
+    )
+
     composite_all = comp10m_data.parent / comp10m_data.name.replace('10m', 'ALL')
     with rasterio.open(composite_all, 'w', **profile) as c10mALL:
         c10mALL.write(x_data)
